/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Dict
 *
 * The Initial Developer of the Original Code is Henrik Gemal.
 * Portions created by the Initial Developer are Copyright (C) 2002-2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Henrik Gemal <dict@gemal.dk> http://gemal.dk
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// TODO: Don't pollute the global namespace
const DICT_STATE_Start                = 0;
const DICT_STATE_WantClientAck        = 1;
const DICT_STATE_WantDefinition       = 2;
const DICT_STATE_WantDefinitionsDone  = 3;
const DICT_STATE_Done                 = 4;
const DICT_STATE_Error                = 5;

var dictStringBundle;

function Observer(connection, word) {
    this._connection = connection;
    this._word       = word;

    this._bufferedData = '';
    this._state = DICT_STATE_Start;

    this._definitions  = [];
}


Observer.prototype.reportStateMachineError = function ()
{
    throw 
        'Unknown problem encountered when defining "' + this._word + '"\n' +
        'If problem presists, please file a bug report\n';
}

Observer.prototype.popResponseLine = function ()
{
    var responseLinePattern = /^(\d+)\s*(.*)\n/;
    
    var result = this._bufferedData.match(responseLinePattern);
    if (result == null) {
        return false;
    }

    this._responseLine     = result[0];
    this._responseCode     = result[1];
    this._responseLineBody = result[2];
    
    // remove the response line from the buffered input
    var lineEndIndex   = this._bufferedData.indexOf('\n');
    this._bufferedData = this._bufferedData.slice(lineEndIndex + 1);
    return true;
}


Observer.prototype.onStreamDataAvailable = function (
    request, 
    inStream, 
    offset, 
    count
) 
{
    try {
        // read the data from the socket into our buffer
        var newData = this._connection.readData(0, count);

        // strip out carriage returns, since they complicate regexp usage
        this._bufferedData += newData.replace(/\r/g, '');

        // do {
        //   {handle data}
        // } while( {there's enough data to handle} );
        do {
            var previousLength = this._bufferedData.length;

            switch (this._state) {

            case DICT_STATE_Start:
                if (this.popResponseLine()) {
                    
                    if (this._responseCode == 220) {
                        // 220 <text> <capabilities> <msg-id> CRLF
                        // (this is the session init sequence)
                        this._connection.sendData('CLIENT experimental Mozilla plugin\r\n');
                        this._connection.sendData('DEFINE * ' + this._word + '\r\n');
                    
                        this._state = DICT_STATE_WantClientAck;
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;


            case DICT_STATE_WantClientAck:

                if (this.popResponseLine()) {
                    if (this._responseCode == 250) {
                        this._state = DICT_STATE_WantDefinitionsDone;
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICT_STATE_WantDefinitionsDone:

                if (this.popResponseLine()) {
                    if (this._responseCode == 150) {
                        // 150: n definitions returned
                    }
                    else if (this._responseCode == 151) {
                        // 151: <word> <database> <name>
                        // The definition follows this line
                        this._state = DICT_STATE_WantDefinition;
                    }
                    else if (this._responseCode == 250) {

                        // we must have all definitions
                        var features = [
                            'status=no', 
                            'toolbar=no',
                            'menubar=no',
                            'location=no',
                            'screenX=0',
                            'screenY=0',
                            'width=100',
                            'height=100'
                        ];

                        var definitionWindow = window.open(
                            '',
                            '',
                            features.join(','));

                        definitionWindow.document.clear();
                        definitionWindow.document.write('<HTML>');
                        definitionWindow.document.write('<HEAD>');
                        definitionWindow.document.write('<TITLE>');
                        definitionWindow.document.write(this._word);
                        definitionWindow.document.write('</TITLE>');
                        definitionWindow.document.write('</HEAD>');
                        definitionWindow.document.write('<BODY>');
                        definitionWindow.document.write('<TABLE width="450"><TR><TD>');

                        var i;
                        for (i = 0; i < this._definitions.length; i++) {
                            definitionWindow.document.write('<B>');
                            definitionWindow.document.write(this._definitions[i].source);
                            definitionWindow.document.write('</B><BR>\n');
                        
                            var deftext = this._definitions[i].text;

                            // Escape HTML entities
                            deftext = deftext.replace(/&/g,   '&amp;');
                            deftext = deftext.replace(/</g,   '&lt;');
                            
                            // replace newlines with <BR> and whitespace with
                            // non-breaking spaces to preserve formatting.
                            // This is better than using <PRE> because it
                            // allows us to use other mark-up tags.
                            deftext = deftext.replace(/\n/g,  '<BR>');
                            deftext = deftext.replace(/ /g,   '&nbsp;');
                            deftext = deftext.replace(/\t/g,  '&nbsp;&nbsp');

                            // Words that are cross-referenced by WordNet show up 
                            // in braces. e.g., "{word}" and "{word(P)}"
                            deftext = deftext.replace(/{(.*?)(\(.*\))?}/g, '$1');

                            // add newlines after <BR> for readability
                            deftext = deftext.replace(/<BR>/g, '<BR>\n');

                            definitionWindow.document.write('<TT>'+deftext+'</TT><BR>\n');

                            // put a visual divider between this entry and the next
                            if (i + 1 < this._definitions.length) {
                                definitionWindow.document.write('<BR>\n');
                            }
                        }
                        definitionWindow.document.write('</TD></TR></TABLE>');
                        definitionWindow.document.write('</BODY></HTML>');
                        definitionWindow.document.close();

                        // HACK: Figure out how to size the window
                        //       to fit the content
                        definitionWindow.sizeToContent();
                        definitionWindow.innerHeight += 70;

                        // If it's too large to display without scrollbars, shorten it
                        if (600 < definitionWindow.innerHeight) {
                            // limit the height
                            definitionWindow.innerHeight = 600;

                            // make wider to account for the width of the 
                            // vertical scrollbar.  If we don't do this, a 
                            // horizontal scrollbar will appear.
                            definitionWindow.innerWidth += 20;  
                        }

                        this._connection.sendData('QUIT\r\n');
                        this._bufferedData = '';
                        this._state = DICT_STATE_Done;

                        this._connection.disconnect();
                    }
                    else if (this._responseCode == 550) {
                        // 550 invalid database

                        // TODO: Handle this gracefully
                        this.reportStateMachineError();
                    }
                    else if (this._responseCode == 552) {
                        // 552 no match
                        alert('No definitions found for "' + this._word + '"');

                        // TODO: Add a UI for finding a nearest match
                        this._connection.disconnect();
                    }
                    else {
                        this.reportStateMachineError();
                    }
                }
                break;

            case DICT_STATE_WantDefinition:
                
                // 151 <word> <database> CRLF
                // <definition line 1>   CRLF
                // <definition line 2>   CRLF
                // . . . 
                // <definition line n>   CRLF
                // . CRLF
                
                // determine if we have the entire definition
                var definitionSentinal = "\n.\n";
                var definitionEnd = this._bufferedData.indexOf(definitionSentinal);
                if (definitionEnd == -1) {
                    break; // wait for the rest of the definition
                }

                var definition = new Object;

                // 151 <word> <database> CRLF
                var result = this._responseLineBody.match(/^(\S+)\s+\S+\s+(.*)$/);
                if (result != null) {
                    definition.word   = result[1];
                    definition.source = result[2];
                }

                // pop off (and store) the definition
                definition.text = this._bufferedData.slice(0, definitionEnd);
                
                // remove the definition from the buffered input
                this._bufferedData = this._bufferedData.slice(
                    definitionEnd + definitionSentinal.length);
                
                // store the definition
                this._definitions.push(definition);
                this._state = DICT_STATE_WantDefinitionsDone;
                break;

            case DICT_STATE_Done:
                break;

            default:
                reportStateMachineError();
                break;
            }
            
        } while (previousLength != this._bufferedData.length);
    }
    catch (ex) {
        this._connection.disconnect();
        alert(ex);
    }
}

Observer.prototype.onStreamClose = function (
    status
) 
{
}

        
function DefineWord(host, port, word)
{
    try {
        var connection = new CBSConnection();
        
        connection.connect(host, port, (void 0), true, null);
        
        var observer = new Observer(connection, word);
        connection.startAsyncRead(observer);
    }
    catch (ex) {
        alert(ex);
    }
}

function GetSelection() 
{
    var selectedWord = gContextMenu.searchSelected();

    // trim leading and trailing whitespace
    selectedWord = selectedWord.replace(/^\s+/, ''); // trim leading whitespace
    selectedWord = selectedWord.replace(/\s+$/, ''); // trim trailing whitespace

    return selectedWord;
}

        
function DefineSelection()
{
    var host = 'dict.org';
    var port = 2628;

    var selectedWord = GetSelection();

    if (selectedWord.indexOf(' ') != -1) {
        alert(
            "Cannot define selections that span multiple words.\n" +
            "Please narrow your selection to just one word.");
        return;
    }

    if (selectedWord.length == 0) {
        alert("Please select a word to define.\n");
        return;
    }

    DefineWord(host, port, selectedWord);
}


// Every time a new browser window is made dictMenuInit will be called
window.addEventListener("load", dictInit, false);

function dictInit() {

    dictStringBundle = document.getElementById("dict_string_bundle");

    var contextMenu = document.getElementById("contentAreaContextMenu");

    // The first time dictInit is called, contextMenu is null.
    // This test prevents an error from being logged in the JavaScript Console
    if (contextMenu != null) {
        contextMenu.addEventListener(
            "popupshowing",
            onDictPopup,
            false);  
    }
}


// hide the Dictionary Search item when apropriate (use same logic as for web search)
function onDictPopup() {

    var item = document.getElementById('dict-define-selection');
    if (gContextMenu.isTextSelected) {

        var selection = GetSelection();
        
        const LENGTHCUTOFF = 15;

        if (LENGTHCUTOFF < selection.length) {
            // selection is too long to appear in menu.  It must be abbreviated.
            selection = selection.slice(0, LENGTHCUTOFF) + '...';
        }

        var labelTemplate = dictStringBundle.getString("context_searchselect.label");

        item.setAttribute("label", labelTemplate.replace('%S', selection));
        item.hidden = false; // show the menu item
    }
    else {
        item.hidden = true; // hide the menu item
    }
}
